#!/usr/bin/env python3

import json
from datetime import datetime
from pathlib import Path
from secrets import token_hex
from django.core.exceptions import ObjectDoesNotExist
from django.utils.timezone import now

from ..conf import API_TOKEN
from .cfdi_cert import SATCertificate
from .cfdi_xml import CFDI
from .pacs import PACComercioDigital

from ..models import Clients


CURRENT_DIR = Path(__file__).resolve().parent
CADENA = 'xslt/cadena.xslt'


def validate_token(token):
    return token == API_TOKEN


def validate_client(post, files):
    rfc = post.get('rfc', '')
    if not rfc:
        msg = 'El RFC es requerido'
        return False, msg

    contra = post.get('password', '')
    if not contra:
        msg = 'La contraseña es requerida'
        return False, msg

    file_cer = files.get('cer', b'')
    if not file_cer:
        msg = 'El archivo CER es requerido'
        return False, msg

    file_key = files.get('key', b'')
    if not file_key:
        msg = 'El archivo KEY es requerido'
        return False, msg

    cert = SATCertificate(file_cer.read(), file_key.read(), contra)
    if not cert.is_valid:
        return False, cert.error

    if cert.is_fiel:
        msg = 'El certificado es FIEL'
        return False, msg

    if rfc.upper() != cert.rfc:
        msg = 'El certificado no corresponde al RFC'
        return False, msg


    data = {
        'rfc': rfc,
        'token': token_hex(32),
        'key': cert.key_enc,
        'cer': cert.cer,
        'serial_number': cert.serial_number,
        'date_from': cert.not_before,
        'date_to': cert.not_after,
    }
    return True, data


def validate_cfdi(post):
    data = json.loads(post)
    rfc = data['emisor']['Rfc']

    try:
        obj = Clients.objects.get(rfc=rfc)
    except ObjectDoesNotExist:
        msg = 'Emisor no existe'
        return False, msg

    if obj.date_to < now():
        msg = 'Certificado vencido'
        return False, msg

    if rfc != obj.rfc:
        msg = 'Documento no corresponde al RFC del emisor'
        return False, msg

    data = {'cfdi': data, 'emisor': obj}

    return True, data


def send_to_pac(data):
    pac = PACComercioDigital()
    xml = pac.stamp(data)
    if pac.error:
        print(pac.error)
        return '', pac.error

    return xml, ''


def send_stamp(data):
    msg = ''
    emisor = data['emisor']
    path_xslt = CURRENT_DIR / CADENA

    cert = SATCertificate(emisor.cer, emisor.key)

    cfdi = CFDI()
    try:
        cfdi.make_xml(data['cfdi'])
        cfdi.stamp(cert, path_xslt)
    except Exception as e:
        return '', str(e)

    xml, msg = send_to_pac(cfdi.xml)

    return xml, msg
